import copy
import math
import os
from collections import OrderedDict, defaultdict

import cloudpickle as pkl
import matplotlib.pyplot as plt
import mxnet as mx
from PIL import Image
from mxnet.gluon.data.vision import transforms

from .metrics import get_metric_instance
from .nets import get_network
from .utils import *
from ..utils import collect_params, update_params
from autogluon.core import AutoGluonObject
from autogluon.core.utils import save, load, tqdm
from autogluon.core.task.base import BasePredictor


__all__ = ['Classifier']


class Classifier(BasePredictor):
    """Trained Image Classifier returned by fit() that can be used to make predictions on new images.

    Attributes
    ----------

    Examples
    --------
    >>> import autogluon.core as ag
    >>> from autogluon.vision import ImagePredictor as task
    >>> dataset = task.Dataset(train_path='data/train',
    >>>                        test_path='data/test')
    >>> classifier = task.fit(dataset,
    >>>                       nets=ag.space.Categorical['resnet18_v1', 'resnet34_v1'],
    >>>                       time_limits=time_limits,
    >>>                       ngpus_per_trial=1,
    >>>                       num_trials = 4)
    >>> image = 'data/test/BabyShirt/BabyShirt_323.jpg'
    >>> ind, prob = classifier.predict(image)
    """

    def __init__(self, model, results, eval_func, scheduler_checkpoint,
                 args, ensemble=0, format_results=True, **kwargs):
        self.model = model
        self.eval_func = eval_func
        self.results = self._format_results(results) if format_results else results
        self.scheduler_checkpoint = scheduler_checkpoint
        self.args = args
        self.ensemble = ensemble

    @classmethod
    def load(cls, checkpoint):
        """Load trained Image Classifier from directory specified by `checkpoint`.
        """
        state_dict = load(checkpoint)
        args = state_dict['args']
        results = pkl.loads(state_dict['results'])
        eval_func = state_dict['eval_func']
        scheduler_checkpoint = state_dict['scheduler_checkpoint']
        model_params = state_dict['model_params']
        ensemble = state_dict['ensemble']

        if ensemble <= 1:
            model_args = copy.deepcopy(args)
            model_args.update(results['best_config'])
            model = get_network(args.net, num_classes=results['num_classes'], ctx=mx.cpu(0))
            update_params(model, model_params)
        else:
            raise NotImplemented
        return cls(model, results, eval_func, scheduler_checkpoint, args,
                   ensemble, format_results=False)

    def state_dict(self, destination=None):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        model_params = collect_params(self.model)
        destination['model_params'] = model_params
        destination['eval_func'] = self.eval_func
        destination['results'] = pkl.dumps(self.results)
        destination['scheduler_checkpoint'] = self.scheduler_checkpoint
        destination['args'] = self.args
        destination['ensemble'] = self.ensemble
        return destination

    def save(self, checkpoint):
        """ Save image classifier to folder specified by `checkpoint`.
        """
        state_dict = self.state_dict()
        save(state_dict, checkpoint)

    def predict(self, X, input_size=224, crop_ratio=0.875, set_prob_thresh=0.001, plot=False):
        """Predict class-index and associated class probability for each image in a given dataset (or just a single image). 
        
        Parameters
        ----------
        X : str or :class:`autogluon.task.ImagePredictor.Dataset` or list of `autogluon.task.ImagePredictor.Dataset`
            If str, should be path to the input image (when we just want to predict on single image).
            If class:`autogluon.task.ImagePredictor.Dataset`, should be dataset of multiple images in same format as training dataset.
            If list of `autogluon.task.ImagePredictor.Dataset`, should be a set of test dataset with different scales of origin images.
        input_size : int
            Size of the images (pixels).
        plot : bool
            Whether to plot the image being classified.
        set_prob_thresh: float
            Results with probability below threshold are set to 0 by default.

        Examples
        --------
        >>> import autogluon.core as ag
        >>> from autogluon.vision import ImagePredictor as task
        >>> train_data = task.Dataset(train_path='~/data/train')
        >>> classifier = task.fit(train_data,
        >>>                       nets=ag.space.Categorical['resnet18_v1', 'resnet34_v1'],
        >>>                       time_limits=600, ngpus_per_trial=1, num_trials=4)
        >>> test_data = task.Dataset('~/data/test', train=False)
        >>> class_index, class_probability = classifier.predict('example.jpg')
        """

        input_size = self.model.input_size if hasattr(self.model, 'input_size') else input_size
        resize = int(math.ceil(input_size / crop_ratio))

        transform_size = transforms.Compose([
            transforms.Resize(resize),
            transforms.CenterCrop(input_size),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ])

        def predict_img(img, ensemble=False):
            proba = self.predict_proba(img)
            if ensemble:
                return proba
            else:
                ind = mx.nd.argmax(proba, axis=1).astype('int')
                idx = mx.nd.stack(mx.nd.arange(proba.shape[0], ctx=proba.context), ind.astype('float32'))
                probai = mx.nd.gather_nd(proba, idx)
                return ind, probai, proba

        def avg_prediction(different_dataset, threshold=0.001):
            result = defaultdict(list)
            inds, probas, probals_all = [], [], []
            for i in range(len(different_dataset)):
                for j in range(len(different_dataset[0])):
                    result[j].append(different_dataset[i][j])

            for c in result.keys():
                proba_all = sum([*result[c]]) / len(different_dataset)
                proba_all = (proba_all >= threshold) * proba_all
                ind = mx.nd.argmax(proba_all, axis=1).astype('int')
                idx = mx.nd.stack(mx.nd.arange(proba_all.shape[0], ctx=proba_all.context), ind.astype('float32'))
                proba = mx.nd.gather_nd(proba_all, idx)
                inds.append(ind.asscalar())
                probas.append(proba.asnumpy())
                probals_all.append(proba_all.asnumpy().flatten())
            return inds, probas, probals_all

        def predict_imgs(X):
            if isinstance(X, list):
                different_dataset = []
                for i, x in enumerate(X):
                    proba_all_one_dataset = []
                    tbar = tqdm(range(len(x.items)))
                    for j, x_item in enumerate(x):
                        tbar.update(1)
                        proba_all = predict_img(x_item[0], ensemble=True)
                        tbar.set_description('ratio:[%d],The input picture [%d]' % (i, j))
                        proba_all_one_dataset.append(proba_all)
                    different_dataset.append(proba_all_one_dataset)
                inds, probas, probals_all = avg_prediction(different_dataset, threshold=set_prob_thresh)
            else:
                inds, probas, probals_all = [], [], []
                tbar = tqdm(range(len(X.items)))
                for i, x in enumerate(X):
                    tbar.update(1)
                    ind, proba, proba_all = predict_img(x[0])
                    tbar.set_description(
                        'The input picture [%d] is classified as [%d], with probability %.2f ' %
                        (i, ind.asscalar(), proba.asscalar())
                    )
                    inds.append(ind.asscalar())
                    probas.append(proba.asnumpy())
                    probals_all.append(proba_all.asnumpy().flatten())
            return inds, probas, probals_all

        if isinstance(X, str) and os.path.isfile(X):
            img = mx.image.imread(filename=X)
            if plot:
                plt.imshow(img.asnumpy())
                plt.show()

            img = transform_size(img)
            return predict_img(img)

        if isinstance(X, AutoGluonObject):
            X = X.init()
            return predict_imgs(X)

        if isinstance(X, list) and len(X) > 1:
            X_group = []
            for X_item in X:
                X_item = X_item.init()
                X_group.append(X_item)
            return predict_imgs(X_group)

    @staticmethod
    def loader(path):
        with open(path, 'rb') as f:
            img = Image.open(f)
            return img.convert('RGB')

    def predict_proba(self, X):
        """Produces predicted class probabilities for a given image.
        """
        pred = self.model(X.expand_dims(0))
        return mx.nd.softmax(pred)

    def evaluate(self, dataset, input_size=224, ctx=[mx.cpu()]):
        """Evaluate predictive performance of trained image classifier using given test data.
        
        Parameters
        ----------
        dataset : :class:`autogluon.task.ImagePredictor.Dataset`
            The dataset containing test images (must be in same format as the training dataset).
        input_size : int
            Size of the images (pixels).
        ctx : List of mxnet.context elements.
            Determines whether to use CPU or GPU(s), options include: `[mx.cpu()]` or `[mx.gpu()]`.
        
        Examples
        --------
        >>> import autogluon.core as ag
        >>> from autogluon.vision import ImagePredictor as task
        >>> train_data = task.Dataset(train_path='~/data/train')
        >>> classifier = task.fit(train_data,
        >>>                       nets=ag.space.Categorical['resnet18_v1', 'resnet34_v1'],
        >>>                       time_limits=600, ngpus_per_trial=1, num_trials = 4)
        >>> test_data = task.Dataset('~/data/test', train=False)
        >>> test_acc = classifier.evaluate(test_data)
        """
        args = self.args
        net = self.model
        batch_size = args.batch_size * max(len(ctx), 1)
        metric = get_metric_instance(args.metric)
        input_size = net.input_size if hasattr(net, 'input_size') else input_size

        test_data, _, batch_fn, _ = get_data_loader(dataset, input_size, batch_size, args.num_workers, True, None)
        tbar = tqdm(test_data)
        for batch in tbar:
            self.eval_func(net, batch, batch_fn, metric, ctx)
            _, test_reward = metric.get()
            tbar.set_description('{}: {}'.format(args.metric, test_reward))
        _, test_reward = metric.get()
        return test_reward

    def evaluate_predictions(self, y_true, y_pred):
        raise NotImplementedError  # TODO



import logging
import mxnet as mx
from mxnet import gluon, init
from mxnet.gluon import nn
import autogluon.core as ag
from autogluon.core import AutoGluonObject

from autogluon.extra.model_zoo.model_zoo import get_model

logger = logging.getLogger(__name__)

__all__ = ['Ensemble', 'get_network', 'auto_suggest_network']

class Ensemble(object):
    def __init__(self, model_list):
        self.model_list = model_list

    def __call__(self, *inputs):
        outputs = [model(*inputs) for model in self.model_list]
        output = outputs[0].exp()
        for i in range(1, len(outputs)):
            output += outputs[i].exp()

        output /= len(outputs)
        return output.log()

class Identity(mx.gluon.HybridBlock):
    def hybrid_forward(self, F, x):
        return x

class ConvBNReLU(mx.gluon.HybridBlock):
    def __init__(self, in_channels, channels, kernel, stride):
        super().__init__()
        padding = (kernel - 1) // 2
        self.conv = nn.Conv2D(channels, kernel, stride, padding, in_channels=in_channels)
        self.bn = nn.BatchNorm(in_channels=channels)
        self.relu = nn.Activation('relu')
    def hybrid_forward(self, F, x):
        return self.relu(self.bn(self.conv(x)))

class ResUnit(mx.gluon.HybridBlock):
    def __init__(self, in_channels, channels, hidden_channels, kernel, stride):
        super().__init__()
        self.conv1 = ConvBNReLU(in_channels, hidden_channels, kernel, stride)
        self.conv2 = ConvBNReLU(hidden_channels, channels, kernel, 1)
        if in_channels == channels and stride == 1:
            self.shortcut = Identity()
        else:
            self.shortcut = nn.Conv2D(channels, 1, stride, in_channels=in_channels)
    def hybrid_forward(self, F, x):
        return self.conv2(self.conv1(x)) + self.shortcut(x)

@ag.func()
def mnist_net():
    mnist_net = gluon.nn.Sequential()
    mnist_net.add(ResUnit(1, 8, hidden_channels=8, kernel=3, stride=2))
    mnist_net.add(ResUnit(8, 8, hidden_channels=8, kernel=5, stride=2))
    mnist_net.add(ResUnit(8, 16, hidden_channels=8, kernel=3, stride=2))
    mnist_net.add(nn.GlobalAvgPool2D())
    mnist_net.add(nn.Flatten())
    mnist_net.add(nn.Activation('relu'))
    mnist_net.add(nn.Dense(10, in_units=16))
    return mnist_net

def auto_suggest_network(dataset, net):
    if isinstance(dataset, str):
        dataset_name = dataset
    elif isinstance(dataset, AutoGluonObject):
        if 'name' in dataset.kwargs and dataset.kwargs['name'] is not None:
            dataset_name = dataset.kwargs['name']
        else:
            return net
    else:
        return net
    dataset_name = dataset_name.lower()
    if 'mnist' in dataset_name:
        if isinstance(net, str) or isinstance(net, ag.Categorical):
            net = mnist_net()
            logger.info('Auto suggesting network net for dataset {}'.format(net, dataset_name))
            return net
    elif 'cifar' in dataset_name:
        if isinstance(net, str):
            if 'cifar' not in net:
                net = 'cifar_resnet20_v1'
        elif isinstance(net, ag.Categorical):
            newdata = []
            for x in net.data:
                if 'cifar' in x:
                    newdata.append(x)
            net.data = newdata if len(newdata) > 0 else ['cifar_resnet20_v1', 'cifar_resnet56_v1']
        logger.info('Auto suggesting network net for dataset {}'.format(net, dataset_name))
        return net

def get_network(net, **kwargs):
    if type(net) == str:
        net = get_built_in_network(net, **kwargs)
    else:
        net.initialize(ctx=kwargs['ctx'])
    return net

def get_built_in_network(name, *args, **kwargs):
    def _get_finetune_network(model_name, num_classes, ctx, **kwargs):
        kwargs['pretrained'] = True
        finetune_net = get_model(model_name, **kwargs)
        # change the last fully connected layer to match the number of classes
        with finetune_net.name_scope():
            if hasattr(finetune_net, 'output'):
                finetune_net.output = gluon.nn.Dense(num_classes)
                finetune_net.output.initialize(init.Xavier(), ctx=ctx)
            elif hasattr(finetune_net, '_fc'):
                finetune_net._fc = gluon.nn.Dense(num_classes)
                finetune_net._fc.initialize(init.Xavier(), ctx=ctx)
            else:
                assert hasattr(finetune_net, 'fc')
                finetune_net.fc = gluon.nn.Dense(num_classes)
                finetune_net.fc.initialize(init.Xavier(), ctx=ctx)
        # initialize and context
        finetune_net.collect_params().reset_ctx(ctx)
        # finetune_net.load_parameters(opt.resume_params, ctx=context, cast_dtype=True)
        finetune_net.hybridize()
        return finetune_net

    def _get_cifar_network(name, num_classes, ctx=mx.cpu(), *args, **kwargs):
        name = name.lower()
        assert 'cifar' in name
        net = get_model(name, *args, **kwargs)
        net.initialize(ctx=ctx)
        return net

    name = name.lower()
    if 'cifar' in name:
        return _get_cifar_network(name, *args, **kwargs)
    else:
        return _get_finetune_network(name, *args, **kwargs)



import logging
import math
import os
import platform
import sys
import warnings

import numpy as np
from PIL import Image
from mxnet import gluon, nd
from mxnet import recordio
from mxnet.gluon.data import RecordFileDataset
from mxnet.gluon.data.vision import ImageFolderDataset as MXImageFolderDataset
from mxnet.gluon.data.vision import ImageRecordDataset, transforms

from autogluon.core import *
from ..utils import get_data_rec
from ..utils.pil_transforms import *

_is_osx = platform.system() == "Darwin"

__all__ = [
    'get_dataset',
    'get_built_in_dataset',
    'ImageFolderDataset',
    'RecordDataset',
    'NativeImageFolderDataset']

logger = logging.getLogger(__name__)

built_in_datasets = [
    'mnist',
    'cifar',
    'cifar10',
    'cifar100',
    'imagenet',
    'fashionmnist',
]


class _TransformFirstClosure(object):
    """Use callable object instead of nested function, it can be pickled."""

    def __init__(self, fn):
        self._fn = fn

    def __call__(self, x, *args):
        if args:
            return (self._fn(x),) + args
        return self._fn(x)


def generate_transform(train, resize, _is_osx, input_size, jitter_param):
    if _is_osx:
        # using PIL to load image (slow)
        if train:
            transform = Compose(
                [
                    RandomResizedCrop(input_size),
                    RandomHorizontalFlip(),
                    ColorJitter(0.4, 0.4, 0.4),
                    ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
                ]
            )
        else:
            transform = Compose(
                [
                    Resize(resize),
                    CenterCrop(input_size),
                    ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
                ]
            )
    else:
        if train:
            transform = transforms.Compose(
                [
                    transforms.RandomResizedCrop(input_size),
                    transforms.RandomFlipLeftRight(),
                    transforms.RandomColorJitter(
                        brightness=jitter_param,
                        contrast=jitter_param,
                        saturation=jitter_param
                    ),
                    transforms.RandomLighting(0.1),
                    transforms.ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
                ]
            )
        else:
            transform = transforms.Compose(
                [
                    transforms.Resize(resize),
                    transforms.CenterCrop(input_size),
                    transforms.ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
                ]
            )
    return transform


@func()
def get_dataset(path=None, train=True, name=None,
                input_size=224, crop_ratio=0.875, jitter_param=0.4, scale_ratio_choice=[],
                *args, **kwargs):
    """ Method to produce image classification dataset for AutoGluon, can either be a
    :class:`ImageFolderDataset`, :class:`RecordDataset`, or a
    popular dataset already built into AutoGluon ('mnist', 'cifar10', 'cifar100', 'imagenet').

    Parameters
    ----------
    name : str, optional
        Which built-in dataset to use, will override all other options if specified.
        The options are ('mnist', 'cifar', 'cifar10', 'cifar100', 'imagenet')
    train : bool, default = True
        Whether this dataset should be used for training or validation.
    path : str
        The training data location. If using :class:`ImageFolderDataset`,
        image folder`path/to/the/folder` should be provided.
        If using :class:`RecordDataset`, the `path/to/*.rec` should be provided.
    input_size : int
        The input image size.
    crop_ratio : float
        Center crop ratio (for evaluation only)
    scale_ratio_choice: list
        List of crop_ratio, only in the test dataset, the set of scaling ratios obtained is scaled to the original image, and then cut a fixed size (input_size) and get a set of predictions for averaging.

    Returns
    -------
    Dataset object that can be passed to `task.fit()`, which is actually an :class:`autogluon.space.AutoGluonObject`.
    To interact with such an object yourself, you must first call `Dataset.init()` to instantiate the object in Python.
    """

    resize = int(math.ceil(input_size / crop_ratio))
    transform = generate_transform(train, resize, _is_osx, input_size, jitter_param)

    if isinstance(name, str) and name.lower() in built_in_datasets:
        return get_built_in_dataset(name, train=train, input_size=input_size, *args, **kwargs)

    if '.rec' in path:
        dataset = RecordDataset(
            path,
            *args,
            transform=_TransformFirstClosure(transform),
            **kwargs
        )
    elif _is_osx:
        dataset = ImageFolderDataset(path, transform=transform, *args, **kwargs)
    elif not train:
        if not scale_ratio_choice:
            dataset = TestImageFolderDataset(
                path,
                *args,
                transform=_TransformFirstClosure(transform),
                **kwargs
            )
        else:
            dataset = []
            for i in scale_ratio_choice:
                resize = int(math.ceil(input_size / i))
                dataset_item = TestImageFolderDataset(
                    path,
                    *args,
                    transform=_TransformFirstClosure(
                        generate_transform(train, resize, _is_osx, input_size, jitter_param)
                    ),
                    **kwargs
                )
                dataset.append(dataset_item.init())

    elif 'label_file' in kwargs:
        dataset = IndexImageDataset(
            path,
            transform=_TransformFirstClosure(transform),
            *args,
            **kwargs
        )
    else:
        dataset = NativeImageFolderDataset(
            path,
            *args,
            transform=_TransformFirstClosure(transform),
            **kwargs
        )

    if not scale_ratio_choice:
        dataset = dataset.init()
    return dataset


@obj()
class IndexImageDataset(MXImageFolderDataset):
    """A image classification dataset with a CVS label file
       Each sample is an image and its corresponding label.

    Parameters
    ----------
    root : str
        Path to the image folder.
    indexfile : str
        Local path to the csv index file. The CSV should have two collums
        1. image name (e.g. xxxx or xxxx.jpg)
        2. label name or index (e.g. aaa or 1)
    gray_scale : False
        If True, always convert images to greyscale.
        If False, always convert images to colored (RGB).
    transform : function, default None
        A user defined callback that transforms each sample.
    """

    def __init__(self, root, label_file, gray_scale=False, transform=None,
                 extension='.jpg'):
        self._root = os.path.expanduser(root)
        self.items, self.synsets = self.read_csv(label_file, root, extension)
        self._flag = 0 if gray_scale else 1
        self._transform = transform

    @staticmethod
    def read_csv(filename, root, extension):
        """The CSV should have two collums
        1. image name (e.g. xxxx or xxxx.jpg)
        2. label name or index (e.g. aaa or 1)
        """

        def label_to_index(label_list, name):
            return label_list.index(name)

        import csv
        label_dict = {}
        with open(filename) as f:
            reader = csv.reader(f)
            for row in reader:
                assert len(row) == 2
                label_dict[row[0]] = row[1]

        if 'id' in label_dict:
            label_dict.pop('id')

        labels = list(set(label_dict.values()))
        samples = [
            (os.path.join(root, f"{k}{extension}"), label_to_index(labels, v))
            for k, v in label_dict.items()
        ]
        return samples, labels

    @property
    def num_classes(self):
        return len(self.synsets)

    @property
    def classes(self):
        return self.synsets

    @property
    def num_classes(self):
        return len(self.synsets)

    @property
    def classes(self):
        return self.synsets


@obj()
class RecordDataset:
    """A dataset wrapping over a RecordIO file containing images.
       Each sample is an image and its corresponding label.

    Parameters
    ----------
    filename : str
        Local path to the .rec file.
    gray_scale : False
        If True, always convert images to greyscale.
        If False, always convert images to colored (RGB).
    transform : function, default None
        A user defined callback that transforms each sample.
    classes : iterable of str, default is None
        User provided class names. If `None` is provide, will use
        a list of increasing natural number ['0', '1', ..., 'N'] by default.
    """

    def __init__(self, filename, gray_scale=False, transform=None, classes=None):
        flag = 0 if gray_scale else 1
        # retrieve number of classes without decoding images
        td = RecordFileDataset(filename)
        s = set([recordio.unpack(td.__getitem__(i))[0].label[0] for i in range(len(td))])
        self._num_classes = len(s)
        if not classes:
            self._classes = [str(i) for i in range(self._num_classes)]
        else:
            if len(self._num_classes) != len(classes):
                warnings.warn('Provided class names do not match data, expected "num_class" is {} '
                              'vs. provided: {}'.format(self._num_classes, len(classes)))
                self._classes = list(classes) + \
                    [str(i) for i in range(len(classes), self._num_classes)]
        self._dataset = ImageRecordDataset(filename, flag=flag)
        if transform:
            self._dataset = self._dataset.transform_first(transform)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def classes(self):
        return self._classes

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        return self._dataset[idx]


@obj()
class NativeImageFolderDataset(MXImageFolderDataset):
    def __init__(self, root, gray_scale=False, transform=None):
        flag = 0 if gray_scale else 1
        super().__init__(root, flag=flag, transform=transform)

    @property
    def num_classes(self):
        return len(self.synsets)

    @property
    def classes(self):
        return self.synsets


@obj()
class TestImageFolderDataset(MXImageFolderDataset):
    def __init__(self, root, gray_scale=False, transform=None):
        flag = 0 if gray_scale else 1
        super().__init__(root, flag=flag, transform=transform)

    def _list_images(self, root):
        self.synsets = []
        self.items = []
        path = os.path.expanduser(root)
        if not os.path.isdir(path):
            raise ValueError('Ignoring %s, which is not a directory.' % path, stacklevel=3)
        for filename in sorted(os.listdir(path)):
            filename = os.path.join(path, filename)
            if os.path.isfile(filename):  # add
                label = len(self.synsets)
                ext = os.path.splitext(filename)[1]
                if ext.lower() not in self._exts:
                    warnings.warn(
                        f'Ignoring {filename} of type {ext}.'
                        f' Only support {", ".join(self._exts)}'
                    )
                    continue
                self.items.append((filename, label))
            else:
                folder = filename
                if not os.path.isdir(folder):
                    raise ValueError(f'Ignoring {path}, which is not a directory.', stacklevel=3)
                label = len(self.synsets)
                for sub_filename in sorted(os.listdir(folder)):
                    sub_filename = os.path.join(folder, sub_filename)
                    ext = os.path.splitext(sub_filename)[1]
                    if ext.lower() not in self._exts:
                        warnings.warn(
                            f'Ignoring {sub_filename} of type {ext}.'
                            f' Only support {", ".join(self._exts)}'
                        )
                        continue
                    self.items.append((sub_filename, label))
                self.synsets.append(label)

    @property
    def num_classes(self):
        return len(self.synsets)

    @property
    def classes(self):
        return self.synsets


@obj()
class ImageFolderDataset(object):
    """A generic data loader where the images are arranged in this way on your local filesystem: ::

        root/dog/a.png
        root/dog/b.png
        root/dog/c.png

        root/cat/x.png
        root/cat/y.png
        root/cat/z.png

    Here, folder-names `dog` and `cat` are the class labels and the images with file-names 'a', `b`, `c` belong to the `dog` class while the others are `cat` images.

    Parameters
    ----------
    root : string
        Root directory path to the folder containing all of the data.
    transform : callable (optional)
        A function/transform that  takes in an PIL image
        and returns a transformed version. E.g, ``transforms.RandomCrop``
    is_valid_file : callable (optional)
        A function that takes path of an Image file
        and check if the file is a valid file (used to check of corrupt files)

    Attributes
    ----------
    classes : list
        List of the class names.
    class_to_idx : dict
        Dict with items (class_name, class_index).
    imgs : list
        List of (image path, class_index) tuples
    """
    _repr_indent = 4
    IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif', '.tiff', '.webp')

    def __init__(self, root, extensions=None, transform=None, is_valid_file=None):
        root = os.path.expanduser(root)
        self.root = root
        extensions = extensions if extensions else self.IMG_EXTENSIONS

        self._transform = transform
        classes, class_to_idx = self._find_classes(self.root)
        samples = self.make_dataset(self.root, class_to_idx, extensions, is_valid_file)
        if len(samples) == 0:
            raise RuntimeError(
                f"Found 0 files in subfolders of:  {self.root} "
                f"\nSupported extensions are:  {','.join(extensions)}"
            )

        self.extensions = extensions

        self.classes = classes
        self.class_to_idx = class_to_idx
        self.samples = samples
        self.targets = [s[1] for s in samples]
        self.imgs = self.samples

    @staticmethod
    def make_dataset(dir, class_to_idx, extensions=None, is_valid_file=None):
        images = []
        dir = os.path.expanduser(dir)
        if not ((extensions is None) ^ (is_valid_file is None)):
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")

        if extensions is not None:

            def is_valid_file(x):
                if not x.lower().endswith(extensions):
                    return False
                valid = True
                try:
                    with open(x, 'rb') as f:
                        Image.open(f)
                except OSError:
                    valid = False
                return valid

        for target in sorted(class_to_idx.keys()):
            d = os.path.join(dir, target)
            if not os.path.isdir(d):
                continue
            for root, _, fnames in sorted(os.walk(d)):
                for fname in sorted(fnames):
                    path = os.path.abspath(os.path.join(root, fname))
                    if is_valid_file(path):
                        item = (path, class_to_idx[target])
                        images.append(item)
        if not class_to_idx:
            for root, _, fnames in sorted(os.walk(dir)):
                for fname in sorted(fnames):
                    path = os.path.abspath(os.path.join(root, fname))
                    if is_valid_file(path):
                        item = (path, 0)
                        images.append(item)
        return images

    @staticmethod
    def loader(path):
        with open(path, 'rb') as f:
            img = Image.open(f)
            return img.convert('RGB')

    @staticmethod
    def _find_classes(dir):
        """Finds the class folders in a dataset.

        Parameters
        ----------
        dir : string
            Root directory path.

        Returns
        -------
        tuple: (classes, class_to_idx)
            where classes are relative to (dir), and class_to_idx is a dictionary.
        """
        if sys.version_info >= (3, 5):
            # Faster and available in Python 3.5 and above
            classes = [d.name for d in os.scandir(dir) if d.is_dir()]
        else:
            classes = [d for d in os.listdir(dir) if os.path.isdir(os.path.join(dir, d))]
        classes.sort()
        class_to_idx = {classes[i]: i for i in range(len(classes))}
        return classes, class_to_idx

    @property
    def num_classes(self):
        return len(self.classes)

    def __getitem__(self, index):
        """
        Parameters
        ----------
        index : int
            Index

        Returns
        ----------
        tuple : (sample, target)
            where target is class_index of the target class.
        """
        path, target = self.samples[index]
        sample = self.loader(path)
        if self._transform is not None:
            sample = self._transform(sample)

        return sample, target

    def __len__(self):
        return len(self.samples)

    def __repr__(self):
        head = "Dataset " + self.__class__.__name__
        body = [f"Number of datapoints: {self.__len__()}"]
        if self.root is not None:
            body.append(f"Root location: {self.root}")
        lines = [head] + [" " * self._repr_indent + line for line in body]
        return '\n'.join(lines)


def get_built_in_dataset(name, train=True, input_size=224, batch_size=256, num_workers=32,
                         shuffle=True, **kwargs):
    """Returns built-in popular image classification dataset based on provided string name ('cifar10', 'cifar100','mnist','imagenet').
    """
    logger.info(f'get_built_in_dataset {name}')
    name = name.lower()
    if name in ('cifar10', 'cifar'):
        import gluoncv.data.transforms as gcv_transforms
        if train:
            transform_split = transforms.Compose(
                [
                    gcv_transforms.RandomCrop(32, pad=4),
                    transforms.RandomFlipLeftRight(),
                    transforms.ToTensor(),
                    transforms.Normalize([0.4914, 0.4822, 0.4465], [0.2023, 0.1994, 0.2010])
                ]
            )
        else:
            transform_split = transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize([0.4914, 0.4822, 0.4465], [0.2023, 0.1994, 0.2010])
                ]
            )
        return gluon.data.vision.CIFAR10(train=train).transform_first(transform_split)
    elif name == 'cifar100':
        import gluoncv.data.transforms as gcv_transforms
        if train:
            transform_split = transforms.Compose(
                [
                    gcv_transforms.RandomCrop(32, pad=4),
                    transforms.RandomFlipLeftRight(),
                    transforms.ToTensor(),
                    transforms.Normalize([0.4914, 0.4822, 0.4465], [0.2023, 0.1994, 0.2010])
                ]
            )
        else:
            transform_split = transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize([0.4914, 0.4822, 0.4465], [0.2023, 0.1994, 0.2010])
                ]
            )
        return gluon.data.vision.CIFAR100(train=train).transform_first(transform_split)
    elif name == 'mnist':
        def transform(data, label):
            return nd.transpose(data.astype(np.float32), (2, 0, 1)) / 255, label.astype(np.float32)

        return gluon.data.vision.MNIST(train=train, transform=transform)
    elif name == 'fashionmnist':
        def transform(data, label):
            return nd.transpose(data.astype(np.float32), (2, 0, 1)) / 255, label.astype(np.float32)

        return gluon.data.vision.FashionMNIST(train=train, transform=transform)
    elif name == 'imagenet':
        # Please setup the ImageNet dataset following the tutorial from GluonCV
        if train:
            rec_file = '/media/ramdisk/rec/train.rec'
            rec_file_idx = '/media/ramdisk/rec/train.idx'
        else:
            rec_file = '/media/ramdisk/rec/val.rec'
            rec_file_idx = '/media/ramdisk/rec/val.idx'
        data_loader = get_data_rec(input_size, 0.875, rec_file, rec_file_idx,
                                   batch_size, num_workers, train, shuffle=shuffle,
                                   **kwargs)
        return data_loader
    else:
        raise NotImplementedError
